/*
    Project: TCGINA
    Version: V1.18
    File   : reg.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include "project.h"

#define countof(arr) (sizeof(arr)/sizeof((arr)[0]))

int GetType(LPTSTR pszType)
{
    const static struct {
        LPTSTR pszName;
        int    iType;
    } Type[] = {
        { TEXT("REG_SZ"), REG_SZ },
        { TEXT("REG_MULTI_SZ"), REG_MULTI_SZ },
        { TEXT("REG_DWORD"), REG_DWORD },
        { NULL, REG_NONE }
    };
    int i;
    for (i = 0; Type[i].pszName != NULL; i++) {
        if (_tcscmp(pszType, Type[i].pszName) == 0) {
            return Type[i].iType;
        }
    }

    return REG_NONE;
}

LPTSTR GetRoot(LPTSTR pszFullKeyName, PHKEY phKey)
{
    const static struct {
        LPTSTR pszName;
        HKEY   hKey;
    } Root[] = {
        { TEXT("HKEY_CLASSES_ROOT"), HKEY_CLASSES_ROOT },
        { TEXT("HKEY_CURRENT_USER"), HKEY_CURRENT_USER },
        { TEXT("HKEY_LOCAL_MACHINE"), HKEY_LOCAL_MACHINE },
        { TEXT("HKEY_USERS"), HKEY_USERS },
        { TEXT("HKCU"), HKEY_CURRENT_USER },
        { TEXT("HKLM"), HKEY_LOCAL_MACHINE },
        { NULL, NULL }
    };
    int i;
    LPTSTR pszKeyName = _tcschr(pszFullKeyName, TEXT('\\'));

    if (pszKeyName == NULL) return NULL;
    *pszKeyName++ = TEXT('\0');

    for (i = 0; Root[i].pszName != NULL; i++) {
        if (_tcscmp(pszFullKeyName, Root[i].pszName) == 0) {
            if (phKey) {
                *phKey = Root[i].hKey;
            }
            return pszKeyName;
        }
    }

    return NULL;
}

BOOL RegSetString(HKEY hRootKey, LPCTSTR pszKeyName, LPCTSTR pszValueName,
  DWORD dwType, LPCTSTR pszValueData)
{
    HKEY hKey;
    BOOL  fResult = FALSE;

    if (RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
      KEY_WRITE, NULL, &hKey, NULL) != ERROR_SUCCESS) {
        return FALSE;
    }
    if (RegSetValueEx(hKey, pszValueName, 0, dwType, (const BYTE*)pszValueData,
      (DWORD)(_tcslen(pszValueData)+1)*sizeof(TCHAR)) == ERROR_SUCCESS) {
        fResult = TRUE;
    }
    RegCloseKey(hKey);

    return fResult;
}

BOOL RegSetInteger(HKEY hRootKey, LPCTSTR pszKeyName, LPCTSTR pszValueName,
  DWORD dwData)
{
    HKEY hKey;
    BOOL fResult = FALSE;

    if (RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
      KEY_WRITE, NULL, &hKey, NULL) != ERROR_SUCCESS) {
        return FALSE;
    }
    if (RegSetValueEx(hKey, pszValueName, 0, REG_DWORD, (const BYTE*)&dwData,
      sizeof(DWORD)) == ERROR_SUCCESS) {
        fResult = TRUE;
    }
    RegCloseKey(hKey);

    return fResult;
}

BOOL RegRemoveKey(HKEY hRootKey, LPCTSTR pszKeyName)
{
    HKEY     hKey;
    LONG     lResult;
    TCHAR    szSubKeyName[MAX_PATH];
    DWORD    dwSize;
    FILETIME ftLastWriteTime;

    if (RegDeleteKey(hRootKey, pszKeyName) == ERROR_SUCCESS) {
        return TRUE;
    }

    lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ, &hKey);
    if (lResult != ERROR_SUCCESS) {
        if (lResult == ERROR_FILE_NOT_FOUND) {
            return TRUE;
        }
        SetLastError(lResult);
        return FALSE;
    }

    dwSize = countof(szSubKeyName);
    while (RegEnumKeyEx(hKey, 0, szSubKeyName, &dwSize, NULL,
      NULL, NULL, &ftLastWriteTime) == ERROR_SUCCESS) {
        if (!RegRemoveKey(hKey, szSubKeyName)) {
            DWORD dwLastError = GetLastError();
            RegCloseKey(hKey);
            SetLastError(dwLastError);
            return FALSE;
        }
        dwSize = countof(szSubKeyName);
    }
    RegCloseKey(hKey);

    lResult = RegDeleteKey(hRootKey, pszKeyName) == ERROR_SUCCESS;
    if (lResult == ERROR_SUCCESS) {
        return TRUE;
    }
    SetLastError(lResult);
    return FALSE;
}

BOOL RegRemoveValue(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName)
{
    HKEY hKey;
    LONG lResult;

    lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_WRITE, &hKey);
    if (lResult != ERROR_SUCCESS) {
        SetLastError(lResult);
        return FALSE;
    }

    lResult = RegDeleteValue(hKey, pszValueName);
    if (lResult != ERROR_SUCCESS) {
        RegCloseKey(hKey);
        SetLastError(lResult);
        return FALSE;
    }
    RegCloseKey(hKey);

    return TRUE;
}

LPTSTR ConvertString(LPCTSTR pszSrc, BOOL fMultiString, PDWORD pcbSize)
{
    DWORD  cbSize;
    LPTSTR pszDestBuffer;
    LPTSTR pszDest;

    cbSize = (DWORD)(_tcslen(pszSrc) + 1 + (fMultiString ? 1 : 0)) *
      sizeof(TCHAR);
    pszDestBuffer = (LPTSTR) LocalAlloc(LMEM_FIXED | LMEM_ZEROINIT, cbSize);
    if (pszDestBuffer == NULL) {
        return NULL;
    }
    pszDest = pszDestBuffer;

    while (*pszSrc) {
        if (pszSrc[0] == '&' && pszSrc[1] == '#') {
            LPTSTR psz;
            ULONG ul = _tcstoul(pszSrc+2, &psz, 10);
            if (*psz == ';' && ul < 0x10000)  {
                *pszDest++ = (TCHAR) ul;
                pszSrc = psz + 1;
                continue;
            }
        }
        *pszDest++ = *pszSrc++;
    }
    if (fMultiString) {
        pszDest += 2;
    } else {
        pszDest = pszDestBuffer + _tcslen(pszDestBuffer) + 1;
    }

    if (pcbSize) {
        *pcbSize = (DWORD)((PBYTE)pszDest - (PBYTE)pszDestBuffer);
    }

    return pszDestBuffer;
}

int _tmain(int argc, _TCHAR *argv[])
{
    if (argc < 2) goto display_syntax;

    if (_tcsicmp(argv[1], TEXT("SET")) == 0) {
        HKEY   hRootKey;
        int    iType = GetType(argv[2]);
        LPTSTR pszKeyName = GetRoot(argv[3], &hRootKey);
        HKEY   hKey;
        LONG   lResult;

        if (argc != 6) goto display_syntax;

        if (iType == REG_NONE || pszKeyName == NULL) goto display_syntax;

        lResult = RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
          KEY_WRITE, NULL, &hKey, NULL);
        if (lResult != ERROR_SUCCESS) return lResult;

        if (iType == REG_SZ) {
            DWORD cbValueData;
            LPTSTR pszValueData = ConvertString(argv[5], FALSE, &cbValueData);
            if (pszValueData == NULL) {
                return ERROR_NOT_ENOUGH_MEMORY;
            }
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_SZ,
              (const BYTE*)pszValueData, cbValueData);
            LocalFree(pszValueData);
            if (lResult != ERROR_SUCCESS) return lResult;
        } else if (iType == REG_MULTI_SZ) {
            DWORD cbValueData;
            LPTSTR pszValueData = ConvertString(argv[5], TRUE, &cbValueData);
            if (pszValueData == NULL) {
                return ERROR_NOT_ENOUGH_MEMORY;
            }
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_MULTI_SZ,
              (const BYTE*)pszValueData, cbValueData);
            LocalFree(pszValueData);
            if (lResult != ERROR_SUCCESS) return lResult;
        } else {
            DWORD dwData = _tcstoul(argv[5], NULL, 0);
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_DWORD,
              (const BYTE*)&dwData, sizeof(DWORD));
            if (lResult != ERROR_SUCCESS) return lResult;
        }
        RegCloseKey(hKey);
    } else if (_tcsicmp(argv[1], TEXT("DELETE")) == 0) {
        HKEY   hRootKey;
        LPTSTR pszKeyName = GetRoot(argv[2], &hRootKey);
        if (argc == 3) {
            if (!RegRemoveKey(hRootKey, pszKeyName)) {
                return GetLastError();
            }
        } else if (argc == 4) {
            if (!RegRemoveValue(hRootKey, pszKeyName, argv[3])) {
                return GetLastError();
            }
        } else {
            goto display_syntax;
        }
    } else {
        goto display_syntax;
    }

    return NO_ERROR;

display_syntax:
    _putts(
        TEXT("REG V") TEXT(PROJECT_VERSION_TEXT) TEXT("\n")
        TEXT("\n")
        TEXT("SYNTAX 1: REG SET Type Key Value Data\n")
        TEXT("\n")
        TEXT("  Type  : Value type (REG_SZ, REG_MULTI_SZ or REG_DWORD)\n")
        TEXT("  Key   : Registry key which must start either with\n")
        TEXT("          HKEY_LOCAL_MACHINE or HKEY_CURRENT_USER\n")
        TEXT("  Value : Value name\n")
        TEXT("  Data  : Value data\n")
        TEXT("\n")
        TEXT("  The registry value is correspondingly set.\n")
        TEXT("  The HTML escape sequence \"&#<d>;\" can be used to define Unicode characters,\n")
        TEXT("  where <d> is a placeholder for the decimal Unicode number.\n")
        TEXT("\n")
        TEXT("SYNTAX 2: REG DELETE Key [Value]\n")
        TEXT("\n")
        TEXT("  The registry key or value is deleted.\n")
        TEXT(""));

    return ERROR_INVALID_PARAMETER;
}
